/*
 * deviced
 *
 * Copyright (c) 2012 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>
#include <fcntl.h>

#include "util.h"
#include "core.h"
#include "display-ops.h"
#include "shared/dbus.h"
#include "core/device-notifier.h"
#include "core/devices.h"

#define ON		"on"
#define OFF		"off"

#define SIGNAL_AMBIENT_MODE			"AmbientMode"
#define CLOCK_START			"clockstart"
#define CLOCK_END			"clockend"
#define TIMEOUT_NONE			(-1)
#define AMBIENT_CLOCK_WAITING_TIME	5000 /* ms */

static int ambient_state;
static pid_t ambient_pid;  /* Ambient Clock pid */
static int ambient_condition;
static unsigned int update_count;

void broadcast_ambient_state(int state)
{
	dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						SIGNAL_AMBIENT_MODE,
						g_variant_new("(i)", (state == 0 ? 0 : 1)));
}

int ambient_get_condition(void)
{
	return ambient_condition;
}

int ambient_get_state(void)
{
	return ambient_state;
}

static void ambient_set_condition(keynode_t *key_nodes, void *data)
{
	int state;

	if (key_nodes == NULL) {
		_E("wrong parameter, key_nodes is null");
		return;
	}

	ambient_condition = vconf_keynode_get_bool(key_nodes);
	_I("ambient mode condition is %d", ambient_condition);

	state = (ambient_condition == 0 ? 0 : 1);
	device_notify(DEVICE_NOTIFIER_DISPLAY_AMBIENT_CONDITION, (void *)&state);

	set_dim_state(state);
}

int ambient_set_state(int on)
{
	if (!ambient_condition)
		return 0;

	broadcast_ambient_state(on);

	update_count = 0;

	if (!on)
		ambient_pid = 0;

	_D("AMBIENT is %s", (on ? ON : OFF));

	ambient_state = on;

	device_notify(DEVICE_NOTIFIER_DISPLAY_AMBIENT_STATE, (void *)&ambient_state);

	return 0;
}

void ambient_check_invalid_state(pid_t pid)
{
	if (pid != INTERNAL_LOCK_AMBIENT)
		return;

	if (backlight_ops.get_lcd_power() == DPMS_OFF)
		return;

	if (is_emulator()) {
		/* In emulator, deviced does not turn off the display. */
		if (backlight_ops.get_lcd_power() == DPMS_ON)
			return;
	}

	if (ambient_get_state() == false)
		return;

	_E("Invalid state! Ambient state is change to off!");

	/* If lcd_power is on and ambient state is true
	 * when pm state is changed to sleep,
	 * deviced doesn't get the clock signal.
	 * deviced just turns off lcd in this case. */

	ambient_set_state(false);
	backlight_ops.off(NORMAL_MODE);
}

static void start_clock(void)
{
	if (pm_cur_state == S_NORMAL ||
	    pm_cur_state == S_LCDDIM ||
	    ambient_state == false)
		return;

	pm_lock_internal(INTERNAL_LOCK_AMBIENT,
	    LCD_OFF, STAY_CUR_STATE, AMBIENT_CLOCK_WAITING_TIME);
}

static void end_clock(pid_t pid)
{
	if (pm_cur_state == S_NORMAL ||
	    pm_cur_state == S_LCDDIM ||
	    ambient_state == false)
		return;

	if (update_count == 0)
		broadcast_lcd_off_late(LCD_OFF_LATE_MODE);

	pm_unlock_internal(INTERNAL_LOCK_AMBIENT,
			LCD_OFF, PM_SLEEP_MARGIN);

	update_count++;
	if (update_count == UINT_MAX)
		update_count = 1;
	ambient_pid = pid;

	_I("enter real ambient state by %d",
	    ambient_pid);
}

int ambient_set_screen(char *screen, pid_t pid)
{
	if (!screen)
		return -EINVAL;

	if (!strncmp(screen, CLOCK_START, strlen(CLOCK_START)))
		start_clock();
	else if (!strncmp(screen, CLOCK_END, strlen(CLOCK_END)))
		end_clock(pid);

	return 0;
}

static void ambient_init(void *data)
{
	int ret;

	ret = vconf_get_bool(VCONFKEY_SETAPPL_AMBIENT_MODE_BOOL,
	    &ambient_condition);
	if (ret < 0) {
		_E("Failed to get ambient mode condition ! (%d)", ret);
		ambient_condition = false;
	}
	_I("ambient mode condition is %d", ambient_condition);

	vconf_notify_key_changed(VCONFKEY_SETAPPL_AMBIENT_MODE_BOOL,
	    ambient_set_condition, NULL);
}

static void ambient_exit(void *data)
{
	ambient_set_state(false);
	vconf_ignore_key_changed(VCONFKEY_SETAPPL_AMBIENT_MODE_BOOL,
			ambient_set_condition);
}

static const struct display_ops ambient_ops = {
	.name     = "ambient",
	.init     = ambient_init,
	.exit     = ambient_exit,
};

DISPLAY_OPS_REGISTER(&ambient_ops)

