/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <errno.h>

#include "common.h"
#include "log.h"
#include "dbus.h"

/* -1 is a default timeout value, it's converted to 25*1000 internally. */
#define DBUS_REPLY_TIMEOUT	(-1)

/* basic information */
typedef struct {
	GDBusConnection *conn;
	GBusType bus_type;
	gboolean priv;
	GList *list_names;	/* dbus_name */
	GList *list_object;	/* dbus_object_handle_s */
	pthread_mutex_t mutex;
} dbus_handle_s;

/* path + interfaces */
typedef struct {
	dbus_handle_s *dh;	/* dbus handle */
	const char *path;	/* object path */
	GList *list_ifaces;	/* dbus_interface_s */
} dbus_object_handle_s;
#define get_dh_from_oh(oh) ((dbus_object_handle_s*)oh)->dh

/* global shared bus : system, session */
static dbus_handle_s g_dh[2];

static dbus_handle_s *_dbus_handle_get_connection(GBusType bus_type);

static GBusType g_default_bus_type = G_BUS_TYPE_SYSTEM;
pthread_mutex_t g_mutex = PTHREAD_MUTEX_INITIALIZER;

GBusType dbus_handle_get_default_bus_type(void)
{
	GBusType type;

	pthread_mutex_lock(&g_mutex);
	type = g_default_bus_type;
	pthread_mutex_unlock(&g_mutex);

	return type;
}

dbus_handle_s * _dbus_handle_get_default_connection(void)
{
	return _dbus_handle_get_connection(dbus_handle_get_default_bus_type());
}

#define dbus_handle_lock(handle) do {\
	assert(handle);\
	pthread_mutex_lock(&((handle)->mutex));\
} while (0);

#define dbus_handle_unlock(handle) do {\
	assert(handle);\
	pthread_mutex_unlock(&(handle)->mutex);\
} while (0);

#define dcl_dbus_handle() dbus_handle_s *dh = (dbus_handle_s *)handle;
#define dcl_dbus_handle_null_check() dbus_handle_s *dh = (dbus_handle_s *)handle;\
	if (!dh) {\
	_E("dbus handle is null\n");\
	return 0;\
	}

static GDBusConnection * _get_bus(GBusType bus_type)
{
	GDBusConnection *conn = NULL;
	GError *err = NULL;

	if (bus_type != G_BUS_TYPE_SYSTEM && bus_type != G_BUS_TYPE_SESSION) {
		_E("Wrong bus_type %d", bus_type);
		return NULL;
	}

	conn = g_bus_get_sync(bus_type, NULL, &err);
	if (!conn || err) {
		_E("failed to get bus:type:%d, %s\n", bus_type, err->message);
		g_error_free(err);
		return NULL;
	}

	return conn;
}

static GDBusConnection * _get_bus_private(GBusType bus_type)
{
	GError *err = NULL;
	GDBusConnection *conn = NULL;
	const char * address;

	if (bus_type != G_BUS_TYPE_SYSTEM && bus_type != G_BUS_TYPE_SESSION) {
		_E("Wrong bus_type %d", bus_type);
		return NULL;
	}

	address = g_dbus_address_get_for_bus_sync(bus_type, NULL, &err);
	if (!address || err) {
		_E("failed to get bus address\n");
		g_error_free(err);
		return NULL;
	}

	conn = g_dbus_connection_new_for_address_sync(address,
			(GDBusConnectionFlags) (G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT |
			G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION),
			NULL, /* GDBusAuthObserver */
			NULL,
			&err);
	if (!conn || err) {
		_E("failed to get private bus\n");
		g_error_free(err);
		return NULL;
	}

	return conn;
}

/* ref cout is 1 */
static dbus_handle_s *_dbus_handle_get_connection(GBusType bus_type)
{
	int ibus = bus_type - 1;
	dbus_handle_s *dh = NULL;

	if (bus_type != G_BUS_TYPE_SYSTEM && bus_type != G_BUS_TYPE_SESSION) {
		_E("Unknown bus type %d", bus_type);
		return NULL;
	}
	dh = &g_dh[ibus];

	dbus_handle_lock(dh);

	if (!dh->conn) {
		dh->conn = _get_bus(bus_type);
		dh->priv = FALSE;
		dh->bus_type = bus_type;
		if (!dh->conn)
			dh = NULL;
	}

	dbus_handle_unlock(dh);

	return dh;
}

/* ref cout is 1 */
static dbus_handle_s *_dbus_handle_get_connection_private(GBusType bus_type)
{
	dbus_handle_s * dh;

	if (bus_type != G_BUS_TYPE_SYSTEM && bus_type != G_BUS_TYPE_SESSION) {
		_E("Unknown bus type %d", bus_type);
		return NULL;
	}

	dh = (dbus_handle_s *)calloc(1, sizeof(dbus_handle_s));
	if (!dh) {
		_E("failed to allocate memory for dbus handle");
		return NULL;
	}

	dbus_handle_lock(dh);

	if (!dh->conn) {
		dh->conn = _get_bus_private(bus_type);
		dh->bus_type = bus_type;
		if (!dh->conn)
			goto err;
	}

	dbus_handle_unlock(dh);

	return dh;
err:
	if (dh) {
		dbus_handle_unlock(dh);
		free(dh);
	}
	return NULL;
}


static void _free_func_signal(gpointer data)
{
	_D("free signal subscribe");
}

guint subscribe_dbus_signal(dbus_handle_h handle, const char *path,
			const char *iface, const char *name,
			GDBusSignalCallback cb, void *data,
			destroy_notified free_func)
{
	dcl_dbus_handle();

	if (!dh) {
		dh = _dbus_handle_get_default_connection();
		if (!dh) {
			_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
			return 0;
		}
	}

	if (!dh->conn) {
		_E("connection is null. check bus status");
		return 0;
	}
	return g_dbus_connection_signal_subscribe(dh->conn, NULL, iface, name, path, NULL, G_DBUS_SIGNAL_FLAGS_NONE, cb, data, _free_func_signal);
}

void unsubscribe_dbus_signal(dbus_handle_h handle, guint id)
{
	dcl_dbus_handle();
	if (!dh) {
		dh = _dbus_handle_get_default_connection();
		if (!dh) {
			_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
			return ;
		}
	}

	if (!dh->conn) {
		_E("connection is null. check bus status");
		return ;
	}

	g_dbus_connection_signal_unsubscribe(dh->conn, id);
}

int _check_type_string_is_container(const char *signature)
{
	if (!signature)
		return FALSE;

	switch (signature[0]) {
	case 'a':
	case 'm':
	case 'r':
	case '(':
	case '{':
	case 'v':
		return TRUE;
	default:
		return FALSE;
	}

	return TRUE;
}

GVariant *dbus_handle_make_simple_array(const char *sig, int *param)
{
	GVariantBuilder *builder = NULL;
	GVariant *var = NULL;
	char format[256];
	int i = 0;

	builder = g_variant_builder_new(G_VARIANT_TYPE(sig));
	if (!builder) {
		_E("failed to g_variant_builder_new");
		return NULL;
	}

	while (param[i])
		g_variant_builder_add(builder, "i", param[i++]);

	snprintf(format, sizeof(format) - 1, "(%s)", sig);
	var = g_variant_new(format, builder);
	g_variant_builder_unref(builder);
	return var;
}

/* todo: looks like garbage... */
static GVariant* _append_variant(const char *signature, const char *param[])
{
	char *ch;
	int i;
	int pi;
	int int_type;
	gboolean bool_type;
	unsigned long long int64_type;
	GVariant *ret;
	int len;
	char container[255];// The maximum length of a signature is 255.
	const char *sig = signature;
	GVariantBuilder *builder = NULL;

	if (!signature || !param)
		return 0;

	/* workaround for user fault "(i) != i" but we treat this as same signature */
	/* G_VARIANT_TYPE("si") return NULL */
	/* todo: actually user have to use correct signature */
	if (!_check_type_string_is_container(signature)) {
		snprintf(container, sizeof(container) - 1, "(%s)", signature);
		sig = container;
	}
	if (!g_variant_type_is_container(G_VARIANT_TYPE(sig))) {
		_E("signature (%s) is not container type", signature);
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE(sig));
	len = strlen(sig);
	pi = 0;
	for (ch = (char *)sig, i = 0; i < len; ++i, ++ch) {
		switch (*ch) {
		case '(':
		case ')':
			continue;
		case 'b':
			bool_type = (atoi(param[pi++]) == 0 ? FALSE : TRUE);
			g_variant_builder_add(builder, "b", bool_type);
			break;
		case 'i':
			int_type = atoi(param[pi++]);
			g_variant_builder_add(builder, "i", int_type);
			break;
		case 'u':
			int_type = strtoul(param[pi++], NULL, 10);
			g_variant_builder_add(builder, "u", int_type);
			break;
		case 't':
			int64_type = atoll(param[pi++]);
			g_variant_builder_add(builder, "t", int64_type);
			break;
		case 's':
			g_variant_builder_add(builder, "s", param[pi++]);
			break;
		case 'a':
			++ch;
			switch (*ch) {
			case 'y':
				g_variant_builder_add(builder, "^ay", param[pi++]);
				++i;
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
	}
	ret = g_variant_builder_end(builder);
	g_variant_builder_clear(builder);
	g_variant_builder_unref(builder);

	return ret;
}

GVariant *dbus_handle_method_sync_with_reply(const char *dest, const char *path,
		const char *iface, const char *method,
		const char *signature, const char *param[])
{
	GError *err = NULL;
	GVariant * var = NULL;
	GVariant * ret = NULL;
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return NULL;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return NULL;
	}

	if (signature && param)
		var = _append_variant(signature, param);

	ret = g_dbus_connection_call_sync(dh->conn,
			dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, DBUS_REPLY_TIMEOUT, NULL, &err);
	if (!ret || err) {
		if (err) {
			_E("failed to g_dbus_connection_call_sync:%s", err->message);
			g_error_free(err);
		} else {
			_E("failed to g_dbus_connection_call_sync");
			g_variant_unref(var);
		}
		return NULL;
	}

	return ret;
}

GVariant *dbus_handle_method_sync_with_reply_var(const char *dest,
						const char *path,
						const char *iface,
						const char *method,
						GVariant *var)
{
	GError *err = NULL;
	GVariant * ret = NULL;
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		if (var)
			g_variant_unref(var);
		return NULL;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		if (var)
			g_variant_unref(var);
		return NULL;
	}

	ret = g_dbus_connection_call_sync(dh->conn,
			dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, DBUS_REPLY_TIMEOUT, NULL, &err);
	if (!ret || err) {
		if (err) {
			_E("failed to g_dbus_connection_call_sync:%s", err->message);
			g_error_free(err);
		} else {
			_E("failed to g_dbus_connection_call_sync");
		}
		return NULL;
	}

	return ret;
}

GVariant *dbus_handle_method_with_unix_fd_list_sync_with_reply(const char *dest,
							const char *path,
							const char *iface,
							const char *method,
							const char *signature,
							const char *param[],
							int *in_fdlist,
							int in_size,
							int **out_fdlist,
							int *out_size)
{
	GError *err = NULL;
	GVariant * var = NULL;
	GVariant * ret = NULL;
	dbus_handle_s *dh = NULL;
	GDBusProxy *proxy = NULL;
	GUnixFDList *g_infdlist = NULL;
	GUnixFDList *g_outfdlist = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return NULL;
	}
	if (in_fdlist && in_size == 0) {
		_E("wrong in_fdlist is not null but in_size is 0");
		return NULL;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return NULL;
	}

	if (signature && param)
		var = _append_variant(signature, param);

	proxy = g_dbus_proxy_new_sync(dh->conn, G_DBUS_PROXY_FLAGS_NONE, NULL, dest, path, iface, NULL, &err);
	if (!proxy) {
		if (err) {
			_E("failed to proxy_new_sync(%s)\n", err->message);
			g_error_free(err);
		} else {
			_E("failed to proxy_new_sync\n");
			if (var)
				g_variant_unref(var);
		}
	}

	/* append fd */
	if (in_fdlist) {
		g_infdlist = g_unix_fd_list_new_from_array(in_fdlist, in_size);
		if (!g_infdlist) {
			_E("failed to g_unix_fd_list_new_from_array\n");
			goto out;
		}
		//g_infdlist = g_unix_fd_list_new();
		//if (g_unix_fd_list_append(g_infdlist, in_fdlist[0], &err) < 0) {
	}

	ret = g_dbus_proxy_call_with_unix_fd_list_sync(proxy, method, var, G_DBUS_CALL_FLAGS_NONE, -1,
		g_infdlist, &g_outfdlist, NULL, &err);
	if (!ret || err) {
		if (err) {
			_E("failed to g_dbus_proxy_call_with_unix_fd_list_sync:%s", err->message);
			g_error_free(err);
		} else {
			_E("failed to g_dbus_proxy_call_with_unix_fd_list_sync:");
			if (var)
				g_variant_unref(var);
			if (g_infdlist)
				g_object_unref(g_infdlist);
		}
		goto out;
	}

	/* fds to out array */
	if (g_outfdlist) {
		*out_size = g_unix_fd_list_get_length(g_outfdlist);
		if (*out_size == 0)
			goto out;

		*out_fdlist = g_unix_fd_list_steal_fds(g_outfdlist, NULL);
	}

out:
	if (g_outfdlist)
		g_object_unref(g_outfdlist);
	if (proxy)
		g_object_unref(proxy);

	return ret;
}

GVariant *dbus_handle_method_with_unix_fd_list_sync_with_reply_var(const char *dest,
								const char *path,
								const char *iface,
								const char *method,
								GVariant *param,
								int *in_fdlist,
								int in_size,
								int **out_fdlist,
								int *out_size)
{
	GError *err = NULL;
	GVariant * ret = NULL;
	dbus_handle_s *dh = NULL;
	GDBusProxy *proxy = NULL;
	GUnixFDList *g_infdlist = NULL;
	GUnixFDList *g_outfdlist = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return NULL;
	}
	if (in_fdlist && in_size == 0) {
		_E("wrong in_fdlist is not null but in_size is 0");
		return NULL;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return NULL;
	}

	proxy = g_dbus_proxy_new_sync(dh->conn, G_DBUS_PROXY_FLAGS_NONE, NULL, dest, path, iface, NULL, &err);
	if (!proxy) {
		if (err) {
			_E("failed to proxy_new_sync(%s)\n", err->message);
			g_error_free(err);
		} else {
			_E("failed to proxy_new_sync\n");
			if (param)
				g_variant_unref(param);
		}
		goto out;
	}

	/* append fd */
	if (in_fdlist) {
		g_infdlist = g_unix_fd_list_new_from_array(in_fdlist, in_size);
		if (!g_infdlist) {
			_E("failed to g_unix_fd_list_new_from_array\n");
			goto out;
		}
		//g_infdlist = g_unix_fd_list_new();
		//if (g_unix_fd_list_append(g_infdlist, in_fdlist[0], &err) < 0) {
	}

	/* send message */
	ret = g_dbus_proxy_call_with_unix_fd_list_sync(proxy, method, param, G_DBUS_CALL_FLAGS_NONE, -1,
		g_infdlist, &g_outfdlist, NULL, &err);
	if (!ret || err) {
		if (err) {
			_E("failed to g_dbus_proxy_call_with_unix_fd_list_sync:%s", err->message);
			g_error_free(err);
		} else {
			_E("failed to g_dbus_proxy_call_with_unix_fd_list_sync:");
			if (param)
				g_variant_unref(param);
			if (g_infdlist)
				g_object_unref(g_infdlist);
		}
		goto out;
	}

	/* copy fds to out array */
	if (g_outfdlist) {
		*out_size = g_unix_fd_list_get_length(g_outfdlist);
		if (*out_size == 0)
			goto out;
		*out_fdlist = g_unix_fd_list_steal_fds(g_outfdlist, NULL);
	}
out:
	if (g_outfdlist)
		g_object_unref(g_outfdlist);
	if (proxy)
		g_object_unref(proxy);
	return ret;
}

int dbus_handle_method_sync(const char *dest,
			const char *path,
			const char *iface,
			const char *method,
			const char *signature,
			const char *param[])
{
	int result;
	gboolean result_bool;
	GVariant *reply = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	reply = dbus_handle_method_sync_with_reply(dest, path, iface, method, signature, param);
	if (!reply)
		return -ECOMM;

	if (g_strcmp0("(i)", g_variant_get_type_string(reply)) == 0) {
		g_variant_get(reply, "(i)", &result);
	} else if (g_strcmp0("(b)", g_variant_get_type_string(reply)) == 0) {
		g_variant_get(reply, "(b)", &result_bool);
		result = (int)result_bool;
	} else {
		result = -ENOMSG;
	}

	g_variant_unref(reply);

	return result;
}

int dbus_handle_method_sync_var(const char *dest,
			const char *path,
			const char *iface,
			const char *method,
			GVariant *param)
{
	int result;
	gboolean result_bool;
	GVariant *reply = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	reply = dbus_handle_method_sync_with_reply_var(dest, path, iface, method, param);
	if (!reply)
		return -ECOMM;

	if (g_strcmp0("(i)", g_variant_get_type_string(reply)) == 0) {
		g_variant_get(reply, "(i)", &result);
	} else if (g_strcmp0("(b)", g_variant_get_type_string(reply)) == 0) {
		g_variant_get(reply, "(b)", &result_bool);
		result = (int)result_bool;
	} else {
		result = -ENOMSG;
	}

	g_variant_unref(reply);

	return result;
}

int dbus_handle_method_sync_timeout(const char *dest,
				const char *path,
				const char *iface,
				const char *method,
				const char *signature,
				const char *param[],
				int timeout)
{
	dbus_handle_s *dh = NULL;
	GError *err = NULL;
	GVariant * var = NULL;
	GVariant * reply = NULL;
	int result = 0;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -EPERM;
	}

	if (signature && param)
		var = _append_variant(signature, param);

	reply = g_dbus_connection_call_sync(dh->conn,
			dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, timeout, NULL, &err);
	if (!reply || err) {
		_E("failed to g_dbus_connection_call_sync:%s", err->message);
		g_error_free(err);
		return -1;
	}

	if (g_strcmp0("(i)", g_variant_get_type_string(reply)) == 0)
		g_variant_get(reply, "(i)", &result);
	else
		result = -ENOMSG;

	g_variant_unref(reply);

	return result;
}

int dbus_handle_method_sync_pairs(const char *dest,
				const char *path,
				const char *iface,
				const char *method,
				int num,
				va_list args)
{
	GError *err = NULL;
	GVariant * reply = NULL;
	char *key, *value;
	int ret = 0;
	GVariant *var;
	GVariantBuilder *builder;
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -1;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));

	for (int i = 0 ; i < num ; i = i + 2) {
		key = va_arg(args, char *);
		value = va_arg(args, char *);
		_I("key(%s), value(%s)", key, value);
		g_variant_builder_add(builder, "{ss}", key, value);
	}

	var = g_variant_new("(a{ss})", builder);
	g_variant_builder_unref(builder);

	reply = g_dbus_connection_call_sync(dh->conn,
		dest, path, iface, method,
		var, NULL, G_DBUS_CALL_FLAGS_NONE, DBUS_REPLY_TIMEOUT, NULL, &err);
	if (!reply || err) {
		_E("failed to g_dbus_connection_call_sync");
		return -1;
	}

	if (g_strcmp0("(i)", g_variant_get_type_string(reply)) == 0)
		g_variant_get(reply, "(i)", &ret);
	else
		ret = -ENOMSG;

	g_variant_unref(reply);

	return ret;
}

int dbus_handle_method_async_pairs(const char *dest,
				const char *path,
				const char *iface,
				const char *method,
				int num,
				va_list args)
{
	char *key, *value;
	GVariant *var;
	GVariantBuilder *builder;
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -1;
	}

	// dict
	builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));

	for (int i = 0 ; i < num ; i = i + 2) {
		key = va_arg(args, char *);
		value = va_arg(args, char *);
		_I("key(%s), value(%s)", key, value);
		g_variant_builder_add(builder, "{ss}", key, value);
	}

	var = g_variant_new("(a{ss})", builder);
	g_variant_builder_unref(builder);

	g_dbus_connection_call(dh->conn, dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL,
			NULL,
			NULL);

	return 0;
}

int dbus_handle_method_async(const char *dest,
			const char *path,
			const char *iface,
			const char *method,
			const char *signature,
			const char *param[])
{
	GVariant * var = NULL;
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -1;
	}

	if (signature && param)
		var = _append_variant(signature, param);

	g_dbus_connection_call(dh->conn, dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL,
			NULL,
			NULL);

	return 0;
}

int dbus_handle_method_async_var(const char *dest,
				const char *path,
				const char *iface,
				const char *method,
				GVariant *param)
{
	dbus_handle_s *dh = NULL;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -1;
	}

	g_dbus_connection_call(dh->conn, dest, path, iface, method,
			param, NULL, G_DBUS_CALL_FLAGS_NONE, -1, NULL,
			NULL,
			NULL);

	return 0;
}

/* callback should free gvariant */
static void _cb_pending(GDBusConnection *conn,
			GAsyncResult *res,
			gpointer user_data)
{
	GVariant *reply = NULL;
	GError *err = NULL;
	pending_call_data *data = (pending_call_data *)user_data;

	reply = g_dbus_connection_call_finish(conn, res, &err);
	if (!reply || err) {
		if (err) {
			_E("no message : [%s]", err->message);
			g_error_free(err);
		} else {
			_E("no message");
		}

		if (data && data->func)
			data->func(NULL, data->data, err);
		goto out;
	}

	if (data && data->func)
		data->func(reply, data->data, err);
out:
	if (data)
		free(data);
}

int dbus_handle_method_async_with_reply(const char *dest,
				const char *path,
				const char *iface,
				const char *method,
				const char *signature,
				const char *param[],
				dbus_pending_cb cb,
				int timeout_msec,
				void *data)
{
	dbus_handle_s *dh = NULL;
	pending_call_data *pdata = NULL;
	GVariant * var = NULL;
	int ret = 0;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	if (timeout_msec < -1) {
		_E("wrong timeout %d", timeout_msec);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -EPERM;
	}

	if (signature && param)
		var = _append_variant(signature, param);

	if (cb) {
		pdata = (pending_call_data*)malloc(sizeof(pending_call_data));
		if (!pdata) {
			ret = -ENOMEM;
			goto err;
		}

		pdata->func = cb;
		pdata->data = data;
	}
	g_dbus_connection_call(dh->conn, dest, path, iface, method,
			var, NULL, G_DBUS_CALL_FLAGS_NONE, timeout_msec, NULL,
			(GAsyncReadyCallback)_cb_pending,
			pdata);

	return ret;
err:
	if (var)
		g_variant_unref(var);
	return ret;
}

int dbus_handle_method_async_with_reply_var(const char *dest,
					const char *path,
					const char *iface,
					const char *method,
					GVariant *param,
					dbus_pending_cb cb,
					int timeout_msec,
					void *data)
{
	dbus_handle_s *dh = NULL;
	pending_call_data *pdata = NULL;
	int ret = 0;

	if (!dest || !path || !iface || !method) {
		_E("wrong parameters dest(%s) path(%s) iface(%s) method(%s)", dest, path, iface, method);
		return -1;
	}

	if (timeout_msec < -1) {
		_E("wrong timeout %d", timeout_msec);
		return -1;
	}

	dh = _dbus_handle_get_default_connection();
	if (!dh) {
		_E("failed to get default connection, bustype:%d", (int)dbus_handle_get_default_bus_type());
		return -EPERM;
	}

	if (cb) {
		pdata = (pending_call_data*)malloc(sizeof(pending_call_data));
		if (!pdata) {
			ret = -ENOMEM;
			goto err;
		}

		pdata->func = cb;
		pdata->data = data;
	}
	g_dbus_connection_call(dh->conn, dest, path, iface, method,
			param, NULL, G_DBUS_CALL_FLAGS_NONE, timeout_msec, NULL,
			(GAsyncReadyCallback)_cb_pending,
			pdata);

	return ret;
err:
	if (param)
		g_variant_unref(param);
	return ret;
}
